"""
A module for instantiating sparse tables with log probabilities.
"""

# System imports
import copy
import operator

# Third-party imports
import numpy as np
import scipy.misc
import itertools

# Local imports
from veroku.factors._factor import Factor
from veroku.factors._factor_template import FactorTemplate


# TODO: consider removing some unused functions
class Categorical(Factor):
    """
    A class for instantiating sparse tables with log probabilities.
    """
    def __init__(self, var_names, log_probs_table, cardinalities):
        """
        Construct a SparseLogTable
        :param var_names: (list) The variable names.
        :param probs: (dictionary) A dictionary with assignments (tuples) as keys and log probabilities as values.
                      Missing assignments are assumed to have zero probability.
        """
        # TODO: add check that assignment lengths are consistent with var_names
        # TODO: add check that cardinalities are consistent with assignments
        super().__init__(var_names=var_names)
        self.log_probs_table = copy.deepcopy(log_probs_table)
        self.var_cards = dict(zip(var_names, cardinalities))

    # TODO: improve this to take missing assignments into account. Alternatively: add functionality to sparsify factor when probs turn to 0
    # TODO: Add variable order sorting
    def equals(self, factor):
        if not isinstance(factor, Categorical):
            raise ValueError(f'factor must be of SparseLogTable type but has type {type(factor)}')
        if self.var_names != factor.var_names:
            return False
        for assign, prob in self.log_probs_table.items():
            if assign not in factor.log_probs_table:
                return False
            elif not np.isclose(factor.log_probs_table[assign], prob):
                return False
        return True

    def copy(self):
        """
        Make a copy of this factor.
        :return: (SparseLogTable) The copy.
        """
        return Categorical(var_names=self.var_names.copy(),
                           log_probs_table=copy.deepcopy(self.log_probs_table),
                           cardinalities=self.var_cards.values())

    @staticmethod
    def _get_shared_order_smaller_vars(smaller_vars, larger_vars):
        """
        larger_vars = ['a', 'c', 'd', 'b']
        smaller_vars = ['c', 'e', 'b']
        return ['c', 'b']
        """
        shared_vars = [v for v in smaller_vars if v in larger_vars]
        remaining_smaller_vars = list(set(larger_vars) - set(shared_vars))
        smaller_vars_new_order = shared_vars + remaining_smaller_vars
        return smaller_vars_new_order

    @staticmethod
    def _intersection_has_same_order(larger_vars, smaller_vars):
        """
        Check if the intersection of two lists has the same order in both lists.
        Will return true if either list is empty? SHOULD THIS BE THE CASE?
        """
        indices_of_smaller_in_larger = [larger_vars.index(v) for v in smaller_vars if v in larger_vars]
        if sorted(indices_of_smaller_in_larger) == indices_of_smaller_in_larger:
            return True
        return False

    # TODO: change back to log form
    def marginalise(self, vrs, keep=False):
        """
        Sum out variables from this factor.
        :param vrs: (list) a subset of variables in the factor's scope
        :param keep: Whether to keep or sum out vrs
        :return: (SparseLogTable) the resulting factor.
        """

        vars_to_keep = super().get_marginal_vars(vrs, keep)
        vars_to_sum_out = [v for v in self.var_names if v not in vars_to_keep]
        nested_table, nested_table_vars = Categorical._get_nested_sorted_probs(new_variables_order_outer=vars_to_keep,
                                                                               new_variables_order_inner=vars_to_sum_out,
                                                                               old_variable_order=self.var_names,
                                                                               old_assign_probs=self.log_probs_table)
        result_table = dict()
        for l1_assign, log_probs_table in nested_table.items():
            prob = scipy.misc.logsumexp(list(log_probs_table.values()))
            result_table[l1_assign] = prob

        result_var_cards = copy.deepcopy(self.var_cards)
        for v in vars_to_sum_out:
            del result_var_cards[v]

        return Categorical(var_names=vars_to_keep, log_probs_table=result_table,
                           cardinalities=result_var_cards.values())

    def observe(self, vrs, values):
        """
        Observe variables to have certain values and return reduced table.
        :param vrs: (list) The variables.
        :param values: (tuple or list) Their values
        :return: (SparseLogTable) The resulting factor.
        """

        vars_unobserved = [v for v in self.var_names if v not in vrs]
        nested_table, nested_table_vars = Categorical._get_nested_sorted_probs(new_variables_order_outer=vrs,
                                                                               new_variables_order_inner=vars_unobserved,
                                                                               old_variable_order=self.var_names,
                                                                               old_assign_probs=self.log_probs_table)
        result_table = nested_table[tuple(values)]
        result_var_cards = copy.deepcopy(self.var_cards)
        for v in vrs:
            del result_var_cards[v]

        return Categorical(var_names=vars_unobserved, log_probs_table=result_table,
                           cardinalities=result_var_cards.values())

    def _assert_consistent_cardinalities(self, factor):
        """
        Assert that the variable cardinalities are consistent between two factors.
        :param factor:
        """
        for var in self.var_names:
            if var in factor.var_cards:
                error_msg = f'Error: inconsistent variable cardinalities: {factor.var_cards}, {self.var_cards}'
                assert self.var_cards[var] == factor.var_cards[var], error_msg

    def absorb(self, factor):
        """
        Multiply this factor with another factor and return the result.
        :param factor: The factor to multiply with.
        :rtype factor: Categorical
        :return: The factor product.
        :rtype: Categorical
        """
        if not isinstance(factor, Categorical):
            raise ValueError(f'factor must be of SparseLogTable type but has type {type(factor)}')
        self._assert_consistent_cardinalities(factor)
        result_table, result_vars = Categorical._complex_table_operation(self.var_names, self.log_probs_table,
                                                                         factor.var_names, factor.log_probs_table,
                                                                         operator.add)
        result_var_cards = copy.deepcopy(self.var_cards)
        result_var_cards.update(factor.var_cards)
        return Categorical(var_names=result_vars, log_probs_table=result_table,
                           cardinalities=result_var_cards.values())

    def cancel(self, factor):
        """
        Divide this factor by another factor and return the result.
        :param factor: (SparseLogTable) The factor to divide by.
        :rtype factor: Categorical
        :return: The factor quotient.
        :rtype: Categorical
        """
        if not isinstance(factor, Categorical):
            raise ValueError(f'factor must be of SparseLogTable type but has type {type(factor)}')
        self._assert_consistent_cardinalities(factor)
        result_table, result_vars = Categorical._complex_table_operation(self.var_names, self.log_probs_table,
                                                                         factor.var_names, factor.log_probs_table,
                                                                         operator.sub)
        result_var_cards = copy.deepcopy(self.var_cards)
        result_var_cards.update(factor.var_cards)
        return Categorical(var_names=result_vars, log_probs_table=result_table,
                           cardinalities=result_var_cards.values())

    def argmax(self):
        """
        Get the Categorical assignment (vector value) that maximises the factor potential.
        :return: The argmax assignment.
        :rtype: int list
        """
        return max(self.log_probs_table.items(), key=operator.itemgetter(1))[0]

    @staticmethod
    def _get_nested_sorted_probs(new_variables_order_outer,
                                 new_variables_order_inner,
                                 old_variable_order, old_assign_probs):
        """
        Reorder probs to a new order and sort assignments.
        :params new_variables_order_outer:
        :params new_variables_order_inner:
        :params old_variable_order:
        :params old_assign_probs: A dictionary of assignment and coresponding probabilities.
        Example:
        old_variable_order = [a, b]
        new_variables_order_outer = [b]

          a  b  c   P(a,b)     return:       b    a  c  P(b,a)
        {(0, 0, 0): pa0b0c0                {(0):{(0, 0): pa0b0,
         (0, 1, 0): pa0b1c0                      (1, 0): pa1b0}
         (1, 0, 1): pa1b0c1                 (1):{(0, 1): pa0b1,
         (1, 1, 1): pa1b1c1}                     (1, 1): pa1b1}}
        """
        # TODO: Complete and improve docstring.
        new_variable_order = new_variables_order_outer + new_variables_order_inner
        new_order_indices = [new_variable_order.index(var) for var in old_variable_order]
        new_assign_probs = dict()
        for old_assign_i, old_prob_i in old_assign_probs.items():
            new_row_assignment = [None] * len(old_assign_i)
            for old_i, new_i in enumerate(new_order_indices):
                new_row_assignment[new_i] = old_assign_i[old_i]
            l1_assign = tuple(new_row_assignment[:len(new_variables_order_outer)])
            if l1_assign not in new_assign_probs:
                new_assign_probs[l1_assign] = dict()
            assign_l2 = tuple(new_row_assignment[len(new_variables_order_outer):])
            new_assign_probs[l1_assign][assign_l2] = old_prob_i
        return new_assign_probs, new_variable_order

    @staticmethod
    def _complex_table_operation(vars_a, table_a, vars_b, table_b, func):
        """
        Operate on a pair of tables which can be sparse and have any combination of overlapping or disjoint variable sets.
        :param vars_a:
        :param table_a:
        :param vars_b:
        :param table_b:
        :param func: The function to apply on pairs of corresponding probabilities in the two tables.
        :return:
        """
        # TODO: Complete and improve docstring.
        larger_table = table_a
        smaller_table = table_b
        larger_table_vars = vars_a
        smaller_table_vars = vars_b
        switched = False
        if len(table_a) < len(table_b):
            larger_table = table_b
            smaller_table = table_a
            larger_table_vars = vars_b
            smaller_table_vars = vars_a
            switched = True

        shared_order_smaller_vars = [v for v in larger_table_vars if v in smaller_table_vars]
        remaining_smaller_vars = list(set(smaller_table_vars) - set(shared_order_smaller_vars))

        new_order_smaller_table_vars = remaining_smaller_vars + shared_order_smaller_vars
        smaller_table, smaller_table_vars = Categorical._get_nested_sorted_probs(remaining_smaller_vars,
                                                                                 shared_order_smaller_vars,
                                                                                 smaller_table_vars, smaller_table)
        smaller_table_vars = copy.deepcopy(new_order_smaller_table_vars)

        # use the nested dictionary (of sub-assignment prob dictionaries)
        result_table = dict()
        for assign_l1, l2_table in smaller_table.items():
            result_l2_table = Categorical._basic_table_operation(larger_table_vars, larger_table,
                                                                 smaller_table_vars, l2_table, func)
            for results_assign, prob in result_l2_table.items():
                # TODO: get better solution than this:
                if (func == operator.sub) and switched:
                    prob = -prob
                result_table[assign_l1 + results_assign] = prob
        result_vars = remaining_smaller_vars + larger_table_vars
        return result_table, result_vars

    @staticmethod
    def _basic_table_operation(larger_table_vars,
                               larger_table,
                               smaller_table_vars,
                               smaller_table,
                               _opperator):
        """
        Efficiently perform operations on corresponding (as indicted by the assignments and variables names)
        elements between larger_table_probs and smaller_table_probs. The smaller table variables must only
        contain variables that is also in the larger and must be sorted to have the same order as in the larger
        (although there can be variables in between in the larger). Also both can be sparse.

        Note: the variables will have the same order as larger_table_vars and a new variable name list is therefore not
              returned.

        :param larger_table: A dictionary of assignment tuples and corresponding probabilities.
        :type larger_table: Categorical
        :param smaller_table: A dictionary of assignment tuples and corresponding probabilities.
        :type smaller_table: Categorical

        :Example:
        larger_table_vars = ['a', 'b', 'c']
        larger_table = {(0, 0, 0): 0.5,
                        (0, 1, 1): 0.2,
                        (1, 1, 0): 0.3}
        smaller_table_vars = ['a', c']
        smaller_table = {(0, 0): 0.2,
                        (0, 1): 0.5,
                        (1, 1): 0.1}
        """

        if not Categorical._intersection_has_same_order(larger_table_vars, smaller_table_vars):
            raise ValueError('Variables must have same relative order.')
        shared_indices_in_larger = [larger_table_vars.index(var) for var in smaller_table_vars if var in larger_table_vars]

        result_probs = dict()

        for lt_assign, lt_prob in larger_table.items():
            assign_smaller = tuple([lt_assign[i] for i in shared_indices_in_larger])
            if assign_smaller in smaller_table:
                rt_prob = smaller_table[assign_smaller]
                result_probs[lt_assign] = _opperator(lt_prob, rt_prob)

        return result_probs

    def _apply_to_probs(self, func, include_assignment=False):
        for assign, prob in self.log_probs_table.items():
            if include_assignment:
                self.log_probs_table[assign] = func(prob, assign)
            else:
                self.log_probs_table[assign] = func(prob)

    def normalise(self):
        """
        Return a normalised copy of the factor.
        :return: The normalised factor.
        """

        factor_copy = self.copy()
        logz = scipy.misc.logsumexp(list(factor_copy.log_probs_table.values()))
        for assign, prob in factor_copy.log_probs_table.items():
            factor_copy.log_probs_table[assign] = prob - logz
        return factor_copy

    def kl_divergence(self, factor, normalise_factor=True):
        """
        Get the KL-divergence D_KL(self||factor) between a normalised version of this factor and another factor.
        Reference https://infoscience.epfl.ch/record/174055/files/durrieuThiranKelly_kldiv_icassp2012_R1.pdf, page 1.
        :param factor: The other factor
        :type factor: Gaussian
        :param normalise_factor: Whether or not to normalise the other factor before computing the KL-divergence.
        :type normalise_factor: bool
        :return: The Kullback-Leibler divergence
        :rtype: float
        """
        normalised_self = self.normalise()
        factor_ = factor
        if normalise_factor:
            factor_ = factor.normalise()
        # TODO: check that this is correct. esp with zeroes.
        logPdivQ = normalised_self.cancel(factor_)
        normalised_self._apply_to_probs(np.exp)

        PlogPdivQ_table, _ = Categorical._complex_table_operation(normalised_self.var_names,
                                                                  normalised_self.log_probs_table,
                                                                  logPdivQ.var_names,
                                                                  logPdivQ.log_probs_table,
                                                                  operator.mul)
        kld = np.sum(list(PlogPdivQ_table.values()))
        if kld < 0.0:
            if np.isclose(kld, 0.0, atol=1e-5):
                #  this is fine (numerical error)
                return 0.0
            print('\nnormalise_factor = ', normalise_factor)
            print('self = ')
            self.show()
            print('\nfactor = ')
            factor_.show()
            raise ValueError(f'Negative KLD: {kld}')
        return kld

    def distance_from_vacuous(self):
        """
        Get the Kullback-Leibler (KL) divergence between this factor and a uniform copy of it.
        :return: The KL divergence.
        :rtype: float
        """
        # make uniform copy
        uniform_factor = self.copy()
        cards = list(uniform_factor.var_cards.values())
        uniform_log_prob = -np.log(np.product(cards))
        uniform_factor._apply_to_probs(lambda x: uniform_log_prob)

        return self.kl_divergence(uniform_factor, normalise_factor=False)

    def potential(self, vrs, assignment):
        """
        Get the value of the factor for a specific assignment.
        :param assignment: The assignment
        :return: The value
        """
        assert set(vrs) == set(self.var_names), 'variables (vrs) do not match factor variables.'
        vrs_to_var_names_indices = [self.var_names.index(v) for v in vrs]
        var_names_order_assignments = tuple([assignment[i] for i in vrs_to_var_names_indices])
        return self.log_probs_table[var_names_order_assignments]

    def show(self, exp_log_probs=True):
        """
        Print the factor.
        :param exp_log_probs: Whether or no to exponentiate the log probabilities (to display probabilities instead of
        log-probabilities)
        :type exp_log_probs: bool
        """
        prob_string = 'log(prob)'
        if exp_log_probs:
            prob_string = 'prob'
        print(self.var_names, ' ', prob_string)
        for assignment, log_prob in self.log_probs_table.items():
            prob = log_prob
            if exp_log_probs:
                prob = np.exp(prob)
            print(assignment, ' ', prob)


class CategoricalTemplate(FactorTemplate):

    def __init__(self, log_probs_table, var_templates):
        """
        Create a Categorical factor template.
        :param log_probs_table: The log_probs_table that specifies the assignments and values for the template.
        :type log_probs_table: tuple:float dict
        :param var_templates: A list of formattable strings.
        :type var_templates: str list

        log_probs_table example:
        {(0, 0): 0.1,
         (0, 1): 0.3,
         (1, 0): 0.1,
         (1, 1): 0.5}
        """
        # TODO: Complete and improve docstring.
        super().__init__(var_templates=var_templates)
        self.log_probs_table = copy.deepcopy(log_probs_table)

    def make_factor(self, format_dict=None, var_names=None):
        """
        Make a factor with var_templates formatted by format_dict to create specific var names.
        :param format_dict: The dictionary to be used to format the var_templates strings.
        :type format_dict: str dict
        :return: The instantiated factor.
        :rtype: Categorical
        """
        if format_dict is not None:
            assert var_names is None
            var_names = [vt.format(**format_dict) for vt in self._var_templates]
        return Categorical(log_probs_table=copy.deepcopy(self.log_probs_table),
                           var_names=var_names, cardinalities=self.var_cards.values())


def _reorder_probs(new_variable_order, old_variable_order, old_assign_probs):
    """
    Reorder categorical table variables to a new order and reorder the associated probabilities
    accordingly.
    :param new_variable_order: The new variable order.
    :type new_variable_order: string list
    :param old_variable_order: The old vriable order (corresponding to old_assign_probs)
    :type old_variable_order: string list
    :param old_assign_probs: The assignment probability list
    :type old_assign_probs: string, float nested list

    Example:
        old_variable_order = [a, b]
        new_variable_order = [b, a]

        a b P(a,b)  return    b a P(a,b)
        0 0  pa0b0            0 0  pa0b0
        0 1  pa0b1            0 1  pa1b0
        1 0  pa1b0            1 0  pa0b1
        1 1  pa1b1            1 1  pa1b1
    """
    # TODO This function is old and still uses the list (instead of dictionary) format for probs. Update it.
    new_order_indices = [new_variable_order.index(var) for var in old_variable_order]
    new_assign_probs = []
    for old_assign_prob_i in old_assign_probs:
        old_row_assignment = old_assign_prob_i[0]
        prob = old_assign_prob_i[1]
        new_row_assignment = [None] * len(old_row_assignment)
        for old_i, new_i in enumerate(new_order_indices):
            new_row_assignment[new_i] = old_row_assignment[old_i]
        new_assign_probs.append([tuple(new_row_assignment), prob])

    return sorted(new_assign_probs, key=lambda x: x[0])



from PIL import Image as PIL_Image
import re


def make_sepset_node_name(node_a_name, node_b_name):
    return 'sepset__' + '__'.join(sorted([node_a_name, node_b_name]))


def change_cluster_graph_edge_color(graph, node_a_name, node_b_name, new_color):
    """
    Change the color of an edge in a cluster graph graphviz object. Note that the cluster graph has cluster and sepset nodes.
    The 'edge' of which the color is changed in this function is actually two edges in the graphviz object
    (the edge between node_a and the sepset node and the edge between the sepset node and node_b).
    :param graph: The graphviz graph
    :param node_a_name:
    :param node_b_name:
    :param new_color:
    :return:

    """

    sepset_node_name = make_sepset_node_name(node_a_name, node_b_name)
    change_graph_edge_color(graph=graph, node_a_name=node_a_name, node_b_name=sepset_node_name, new_color=new_color)
    change_graph_edge_color(graph=graph, node_a_name=node_b_name, node_b_name=sepset_node_name, new_color=new_color)


def check_non_overlapping_substring_presence(substring_a, substring_b, string):
    """
    Check that both sub-strings can be found in seperate locations in strings
    Examples:
        substring_a = abc
        substring_b = 123
        string = abc123 -- def
        returns False

        substring_a = abc
        substring_b = 123
        string = abc -- 123
        returns True

    :param substring_a: The one substring to check for.
    :param substring_b: The other substring to check for.
    :param string: The string to check in.
    :return: Result of check for independent string presence.
    """

    if substring_a in substring_b:
        if substring_a not in string:
            return False
        string = string.replace(substring_b, '')
        if substring_a in string:
            return True
    if substring_b in substring_a:
        if substring_b not in string:
            return False
        string = string.replace(substring_a, '')
        if substring_b in string:
            return True
    return (substring_a in string) and (substring_b in string)


def change_graph_edge_color(graph, node_a_name, node_b_name, new_color):
    """
    Change the edge color of the edge between two nodes in a graphviz object.
    :param graph: The graphviz graph
    :param node_a_name: The one node
    :param node_b_name: The other node
    :param new_color: The new color of the edge (i.e 'green', 'blue', 'red')
    :return:
    """

    edge_found = False
    for i, s in enumerate(graph.body):
        if '--' in s:
            if check_non_overlapping_substring_presence(node_a_name, node_b_name, s):
                pattern = f'(\\t?"?{node_a_name}"?\s--\s"?{node_b_name}"?\s\[.*color=)([\S]*\s)(.*)'
                new_s = re.sub(pattern, f'\g<1>{new_color} \g<3>', s)
                # TODO: improve this (bit hacky)
                if new_s == s:
                    pattern = f'(\\t?"?{node_b_name}"?\s--\s"?{node_a_name}"?\s\[.*color=)([\S]*\s)(.*)'
                    new_s = re.sub(pattern, f'\g<1>{new_color} \g<3>', s)

                graph.body[i] = new_s
                edge_found = True
    if not edge_found:
        raise ValueError(
            f'cannot change colour of not existing edge: no edge between node {node_a_name} and node {node_b_name} in graph')


def change_graph_node_color(graph, node_name, new_color):
    """
    Change the fill color of a node in a cluster graph graphviz object.
    :param graph:  The graphviz graph
    :param node_name: The name of the node which colour should be changed.
    :param new_color: The new fill color of the node (i.e 'green', 'blue', 'red')
    :return:
    """
    node_found = False
    for i, s in enumerate(graph.body):
        if '--' not in s:  # not edge
            if node_name in s:
                graph.body[i] = re.sub(f'(\\t?"?{node_name}"?\s\[.*fillcolor=)([\S]*\s)(.*)',f'\g<1>{new_color} \g<3>', s)
                node_found = True
    if not node_found:
        raise ValueError(f'cannot change colour of not existing node: no node {node_name} in graph')


def graph_to_pil_image(graph):
    graph.render('/tmp/test.gv', view=False)
    pil_image = PIL_Image.open('/tmp/test.gv.png')
    pil_image_copy = pil_image.copy()
    pil_image.close()
    return pil_image_copy


def add_message_pass_animation_frames(graph, frames, node_a_name, node_b_name):
    new_color = 'red'
    change_graph_node_color(graph=graph,
                            node_name=node_a_name,
                            new_color=new_color)
    frames.append(graph_to_pil_image(graph))

    change_cluster_graph_edge_color(graph=graph,
                                    node_a_name=node_a_name,
                                    node_b_name=node_b_name,
                                    new_color=new_color)
    frames.append(graph_to_pil_image(graph))

    change_graph_node_color(graph=graph,
                            node_name=node_b_name,
                            new_color=new_color)
    frames.append(graph_to_pil_image(graph))
    change_graph_node_color(graph=graph,
                            node_name=node_a_name,
                            new_color='white')
    frames.append(graph_to_pil_image(graph))

    change_cluster_graph_edge_color(graph=graph,
                                    node_a_name=node_a_name,
                                    node_b_name=node_b_name,
                                    new_color='black')
    frames.append(graph_to_pil_image(graph))
    change_graph_node_color(graph=graph,
                            node_name=node_b_name,
                            new_color='white')
    frames.append(graph_to_pil_image(graph))



"""
A module containing general functions to help with factor operations and transformations.
"""


# TODO: simplify make_square_matrix and make_column vector (maybe renam to make ____ from list - there is already ...
# TODO: ... such a function for msquare matrix). These functions do not need al tie current functionality. ...
# TODO: ... Ensure factor classes still work.

import numpy as np
import matplotlib.pyplot as plt
import copy


def format_list_elements(lst, format_dict):
    formatted_list = [e.format(**format_dict) for e in lst]
    return formatted_list


def get_subset_evidence(all_evidence_dict, subset_vars):
    """
    Select evidence for certain variables only from a evidence_dict and return the variables for which there is evidence
     and the corresponding values.
    :param all_evidence_dict: (dict) The evidence dictionary.
    :param subset_vars: (string list) the subset of variables for which to select evidence from all_evidence_dict.
    :return: evidence variable names and values
    """
    factor_scope_evidence = {v: all_evidence_dict[v] for v in subset_vars if v in all_evidence_dict}
    evidence_vrs = list(factor_scope_evidence.keys())
    evidence_values = list(factor_scope_evidence.values())
    return evidence_vrs, evidence_values


def get_closest_psd_from_symmetric_matrix(mat):
    V, Q = np.linalg.eig(mat)
    Vp = np.maximum(V, 0.0)
    Dp = np.diag(Vp)
    closest_psd_mat = Q.dot(Dp).dot(Q.T)
    return closest_psd_mat


def is_pos_def(matrix):
    """
    Check if matrix is positive definite
    :param matrix: the matrix to check
    :return: (bool) whether the matrix is positive-definite or not
    """
    return np.all(np.linalg.eigvals(matrix) > 0)


def remove_duplicate_values(array_like, tol=0.0):
    """
    Removes duplicate values from list (when tol=0.0) or remove approximately duplicate
    values if tol!=0.0
    """
    unique_values = [array_like[0]]
    for element in array_like:
        element_is_duplicate = False
        for uval in unique_values:
            if abs(uval - element) <= tol:
                element_is_duplicate = True
        if not element_is_duplicate:
            unique_values.append(element)
    return unique_values


def make_list(list_like_object):
    """
    Convert a object to a list
    :param list_like_object: a list like object (i.e string or 1d numpy array)
    :return: the equivalent list
    """
    if not isinstance(list_like_object, list):
        lst = [list_like_object]
        return lst
    return list_like_object


def make_scalar(scalar_like_object):
    """
    Convert a list or numpy array containing a single element to a scalar.
    :param scalar_like_object: a list or numpy array containing a single element (either int or float).
    :return: the int or float contained in the scalar_like_object
    """
    if isinstance(scalar_like_object, np.ndarray):
        if len(scalar_like_object) != 1:
            raise ValueError('array must have length of one to be convertible to a scalar.')
        return scalar_like_object.min()
    if isinstance(scalar_like_object, list):
        if len(scalar_like_object) != 1:
            raise ValueError('list must have length of one to be convertible to a scalar.')
        return scalar_like_object[0]
    if isinstance(scalar_like_object, (int, float)):
        return scalar_like_object
    raise ValueError(f'{type(scalar_like_object)} is not supported')

def make_column_vector(vector_like_object):
    """
    This function converts a vector like object into a standard
    numpy column vector.
    :param vector_like_object: an int, float, list or (one dimensional) array
    :return: a numpy column vector (array)
    """
    array = copy.deepcopy(vector_like_object)
    if isinstance(array, np.ndarray):
        array = array.ravel()
        error_msg = f'Error: array of shape {vector_like_object.shape} cannot be convertes to column vector'
        if len(vector_like_object.shape) > 1:
            assert not all([s > 1 for s in vector_like_object.shape]), error_msg
    else:
        array = np.array(vector_like_object).ravel()
    return np.expand_dims(array, axis=1)


#TODO: simplify this function like the one above
# pylint: disable=inconsistent-return-statements
def list_to_square_matrix(matrix_like_object):
    """
    Make a square matrix from a list.
    :param matrix_like_object: a list containing a single int or float or n lists of length n
    :return: an nxn numpy array.
    """
    len_list = len(matrix_like_object)
    if len_list == 0:
        raise ValueError('cannot make matrix from empty list.')
    if len_list > 1:
        for element in matrix_like_object:
            if not isinstance(element, list):
                raise ValueError('cannot make square matrix from one dimensional list')
            if len(element) != len_list:
                raise ValueError('cannot make square matrix from different length lists')
            # if isinstance(element[0], list):
            #    raise ValueError('cannot make square matrix from 3d list')
        return np.array(matrix_like_object)
    # len_list == 1:
    if isinstance(matrix_like_object[0], (int, float)):
        # matrix_like_object = [float or int]
        return np.array([matrix_like_object])
    if len(matrix_like_object[0]) == 1:
        if isinstance(matrix_like_object[0][0], (int, float)):
            # matrix_like_object = [[float or int]]
            return np.array(matrix_like_object)
    raise ValueError(f'cannot convert matrix_like_object {matrix_like_object} to square matrix')
# pylint: enable=inconsistent-return-statements


# pylint: disable=inconsistent-return-statements
def make_square_matrix(matrix_like_object):
    """
    This function converts a matrix like object into a standard
    numpy square matrix.
    :param matrix_like_object: an int, float, list, list of lists or (square, two dimensional) array
    :return: a numpy square matrix (array)
    """
    if isinstance(matrix_like_object, (int, float)):
        return np.array([[matrix_like_object]])
    if isinstance(matrix_like_object, list):
        return list_to_square_matrix(matrix_like_object)

    if isinstance(matrix_like_object, np.ndarray):
        if len(matrix_like_object.shape) > 2:
            raise ValueError('cannot convert higher (than 2) dimensional tensor to 2d matrix')
        if len(matrix_like_object.shape) == 1:
            return np.expand_dims(matrix_like_object, axis=0)
        assert matrix_like_object.shape[0] == matrix_like_object.shape[1]
        return matrix_like_object.copy()
    raise ValueError(f'cannot convert matrix_like_object {matrix_like_object} to square matrix.')
# pylint: enable=inconsistent-return-statements


def indexed_square_matrix_operation(mat_a, mat_b, var_names_a, var_names_b, operator):
    # TODO: update usages
    """
    A function that performs element wise operations between square matrices where the corresponding indices in the
    different matrices do not necessarily correspond to the same variables, but can be mapped to each other through the
    variable name lists (var_names_a and var_names_b).

    :param mat_a: A nxn numpy array to be used with the operator
    :param mat_b:  A nxn numpy array to be used with the operator
    :param var_names_a: the variable names corresponding to the values in mat_a
    :param var_names_b: the variable names corresponding to the values in mat_b
    :param operator: the operator to be applied to mat_a and mat_b
    :return: the result of the operation

    Example:
        With:
            mat_a =
            [[11, 12, 13],
             [21, 22, 23],
             [31, 32, 33]]
            var_names_a = ['var1', 'var2']
            mat_b =
            [[11 , 13],
             [31,  33]]
            var_names_b = ['var1', 'var3']
            operator_str = '-'
        the function returns:
        [[0,  12, 0],
         [21, 22, 23],
         [0,  32, 0]]
    """

    assert len(mat_a.shape) == 2, 'Error: matrix needs to be 2 dimensional'
    assert len(mat_b.shape) == 2, 'Error: matrix needs to be 2 dimensional'
    assert mat_a.shape[0] == mat_a.shape[1], 'Error: matrix needs to be square'
    assert mat_b.shape[0] == mat_b.shape[1], 'Error: matrix needs to be square'
    if len(var_names_a) != mat_a.shape[0]:
        print('var_names_a = ', len(var_names_a))
        print('mat_a = \n', mat_a.shape)
        raise AssertionError('The number of variables in var_names_a does not match the dimensions of mat_a.')
    if len(var_names_b) != mat_b.shape[0]:
        print('var_names_b = ', len(var_names_b))
        print('mat_b = \n', mat_b.shape)
        raise AssertionError('The number of variables in var_names_b does not match the dimensions of mat_b.')
    # TODO: clean this up (also see todo in indexed_column_vector_operation)
    if set(var_names_b) <= set(var_names_a):
        new_vars = var_names_a
    elif set(var_names_a) <= set(var_names_b):
        new_vars = var_names_b
    else:
        new_vars = list(set(var_names_b + var_names_a))
        new_vars.sort()
    new_dim = len(new_vars)
    new_matrix = np.zeros([new_dim, new_dim])

    for i, var_i_name in enumerate(new_vars):
        for j, var_j_name in enumerate(new_vars):
            val_a = 0.0
            if (var_i_name in var_names_a) and (var_j_name in var_names_a):
                val_a = mat_a[var_names_a.index(var_i_name), var_names_a.index(var_j_name)]
            val_b = 0.0
            if (var_i_name in var_names_b) and (var_j_name in var_names_b):
                val_b = mat_b[var_names_b.index(var_i_name), var_names_b.index(var_j_name)]
            new_matrix[i, j] = operator(val_a, val_b)

    return new_matrix, new_vars


def indexed_column_vector_operation(colvec_a, colvec_b, var_names_a, var_names_b, operator):
    # TODO: modify to match indexed_square_matrix_operation and update usages
    """
    A function that performs element operations between column vectors, where the corresponding indices in the different
    vectors do not necessarily correspond to the same variables, but can be mapped to each other through the variable
    name lists (var_names_a and var_names_b).

    :param colvec_a: A nx1 numpy array to be used with the operator
    :param colvec_b:  A nx1 numpy array to be used with the operator
    :param var_names_a: the variable names corresponding to the values in colvec_a
    :param var_names_b: the variable names corresponding to the values in colvec_b
    :param operator: the operator to be applied to colvec_a and colvec_b
    :return: the result of the operation

    Example:
        With:
            colvec_a =
            [[11],
             [21],
             [31]]
            var_names_a = ['var1', 'var2']
            colvec_b =
            [[11],
             [31]]
            var_names_b = [var1', 'var3']
            operator_str = '-'
        the function returns:
        [[0],
         [21],
         [0]]
    """

    assert len(colvec_a.shape) == 2, 'Error: column vector needs to be 2 dimensional'
    assert len(colvec_b.shape) == 2, 'Error: column vector needs to be 2 dimensional'
    assert colvec_a.shape[1] == 1, 'Error: column vector array cannot have more than one column'
    assert colvec_b.shape[1] == 1, 'Error: column vector array cannot have more than one column'
    assert len(var_names_a) == colvec_a.shape[0], \
        'Error: number of variables in var_names_a does not match the dimension of colvec_a.'
    assert len(var_names_b) == colvec_b.shape[0], \
        'Error: number of variables in var_names_a does not match the dimension of colvec_b.'
    # TODO: clean this up (also see todo in indexed_square_matrix_operation)
    if set(var_names_b) <= set(var_names_a):
        new_vars = var_names_a
    elif set(var_names_a) <= set(var_names_b):
        new_vars = var_names_b
    else:
        new_vars = list(set(var_names_b + var_names_a))
        new_vars.sort()
    new_dim = len(new_vars)
    new_vector = np.zeros([new_dim, 1])

    for i, var_i_name in enumerate(new_vars):
        val_a = 0.0
        if var_i_name in var_names_a:
            val_a = colvec_a[var_names_a.index(var_i_name), 0]
        val_b = 0.0
        if var_i_name in var_names_b:
            val_b = colvec_b[var_names_b.index(var_i_name), 0]
        new_vector[i, 0] = operator(val_a, val_b)

    return new_vector, new_vars


def plot_2d(func, xlim, ylim, xlabel, ylabel, figsize=None):  # pragma: no cover
    """
    Plot a 2d function
    :param xlim: the x limits to plot the function over.
    :param ylim: the y limits to plot the function over.
    :param func: the function to plot
    :param xlim: the x limits over which to plot the function
    :param ylim: the x limits over which to plot the function
    :param xlabel: the x label for the plot
    :param ylabel: the y label for the plot
    :param figsize: the figure size
    """

    x_space = np.linspace(xlim[0], xlim[1], num=100)
    y_space = np.linspace(ylim[0], ylim[1], num=100)

    x_mesh, y_mesh = np.meshgrid(x_space, y_space)
    xx_mesh_array = np.array([x_mesh.ravel(), y_mesh.ravel()]).T
    z_mesh_array = np.array([func(X) for X in xx_mesh_array])
    z_mesh = z_mesh_array.reshape(x_mesh.shape)

    if figsize is not None:
        plt.figure(figsize=figsize)
    color_scale = plt.contourf(x_mesh, y_mesh, z_mesh, levels=30, cmap=plt.cm.viridis)
    plt.colorbar(color_scale, shrink=0.8, extend='both')

    plt.xlabel(xlabel)
    plt.ylabel(ylabel)



"""
This module contains an abstract parent class, defining the minimum functions that all factors should have
"""

from abc import ABCMeta, abstractmethod
import copy

from veroku.factors import _factor_utils


class Factor:
    """
    An abstract parent class.
    """
    __metaclass__ = ABCMeta

    def __init__(self, var_names):
        """
        A super class constructor that should be called from the base class constructor.
        Variable sets explanation:
        #TODO: expand explanation.
        """
        if len(set(var_names)) != len(var_names):
            raise ValueError('duplicate variables in var_names: ', var_names)
        self._var_names = _factor_utils.make_list(var_names)
        self._dim = len(var_names)

    @property
    def var_names(self):
        """
        Get the factor's variable names.
        :return: the var_names parameter.
        """
        return copy.deepcopy(self._var_names)

    @property
    def dim(self):
        """
        Get the factor's dimensionality.
        :return: the dim parameter.
        """
        return self._dim

    def get_marginal_vars(self, vrs, keep=False):
        """
        A helper function (for marginalise) that returns the variables that should be marginalised out (keep=False)
        or kept (keep=True).
        :param vrs: (list) the variables
        :param keep: Whether these variables are to be kept or summed out.
        :return: The variables to be kept or summed out.
        """
        if keep:
            return vrs.copy()
        return list(set(self.var_names) - set(vrs))

    @abstractmethod
    def copy(self):
        """
        An abstract function for copying a factor that should be implemented in the base class.
        :return: a copy of the factor
        """

    @abstractmethod
    def marginalise(self, vrs, keep=False):
        """
        An abstract function for performing factor marginalisation that should be implemented in the base class.
        :return: the resulting marginal factor.
        :param vrs: (list) a subset of variables in the factor's scope
        :param keep: Whether to keep or sum out vrs
        :return: The resulting factor marginal.
        """

    @abstractmethod
    def absorb(self, factor):
        """
        An abstract function for performing factor multiplication that should be implemented in the base class.
        :param factor: The factor to be multiplied with.
        :return: The resulting product
        """

    @abstractmethod
    def cancel(self, factor):
        """
        An abstract function for performing factor division that should be implemented in the base class.
        :param factor: The factor to be divided by.
        :return: The resulting quotient
        """

    @abstractmethod
    def observe(self, vrs, values):
        """
        An abstract function for performing the observation (a.k.a conditioning) operation that should be implemented
        in the base class.
        :return: the resulting reduced factor.
        :param vrs: (list) a subset of variables in the factor's scope
        :param values: The values of vars.
        :return: The resulting reduced factor.
        """

    @abstractmethod
    def show(self):
        """
        An abstract function for printing the parameters of the factor that should be implemented in the base class.
        """

    def __mul__(self, other):
        return self.absorb(other)

    def __div__(self, other):
        return self.cancel(other)
